from setuptools import setup

setup(
    name='tablemap',
    version='0.2.17',
    description='sql not knowing sql',
    long_description='',
    url='https://github.com/nalssee/tablemap.git',
    author='nalssee',
    author_email='jinisrolling@gmail.com',

    license='MIT',

)
